/* $Id$

This file is part of libmsppath
Copyright © 2006-2007  Mikko Rasa, Mikkosoft Productions
Distributed under the LGPL
*/

#include <dirent.h>
#include <errno.h>
#include <msp/core/error.h>
#ifndef WIN32
#include <fnmatch.h>
#else
#include <msp/strings/glob.h>
#endif
#include <msp/strings/utils.h>
#include "path.h"
#include "utils.h"

using namespace std;

namespace Msp {

/**
Fixes the case of the given path to match files / directories on the
filesystem.  Intended to be used in programs that need to interact with
emulated Windows programs.
*/
Path fix_case(const Path &path)
{
	bool found=true;
	Path result;
	for(Path::Iterator i=path.begin(); i!=path.end(); ++i)
	{
		if(!found || *i=="/")
			result/=*i;
		else
		{
			list<string> files;
			if(result.size())
				files=list_files(result);
			else
				files=list_files(".");

			found=false;
			for(list<string>::iterator j=files.begin(); (j!=files.end() && !found); ++j)
				if(!strcasecmp(*j,*i))
				{
					result/=*j;
					found=true;
				}

			if(!found)
				result/=*i;
		}
	}

	return result;
}

void mkdir(const Path &path, int mode)
{
	int err;
#ifdef WIN32
	// The win32 version of this function doesn't take the mode argument.  Go figure.
	(void)mode;
	err=::mkdir(path.str().c_str());
#else
	err=::mkdir(path.str().c_str(), mode);
#endif

	if(err==-1)
		throw SystemError("mkdir failed", errno);
}

/**
Creates a directory and any parent directories if needed.

@param   path  The path to create
@param   mode  Access mode for new directories

@return  0 on success, -1 on error
*/
void mkpath(const Path &path, int mode)
{
	Path p;
	for(Path::Iterator i=path.begin(); i!=path.end(); ++i)
	{
		p/=*i;
#ifdef WIN32
		if(p.size()==1 && is_windows_drive(*i))
			continue;
#endif
		struct stat st;
		int err=stat(p, st);
		if(err==0)
		{
			if(!S_ISDIR(st.st_mode))
				throw Exception("A component exists and is not a directory");
			continue;
		}
		else if(errno!=ENOENT)
			throw SystemError("stat failed", errno);
		else
			mkdir(p, mode);
	}
}

void rmdir(const Path &path)
{
	if(::rmdir(path.str().c_str())==-1)
		throw SystemError("rmdir failed", errno);
}

void rmdirs(const Path &path)
{
	list<string> files=list_files(path);
	for(list<string>::iterator i=files.begin(); i!=files.end(); ++i)
	{
		Path p=path / *i;
		struct stat st=stat(p.str().c_str());
		if(S_ISDIR(st.st_mode))
			rmdirs(p);
		else
			unlink(p);
	}

	rmdir(path);
}

void unlink(const Path &path)
{
	if(::unlink(path.str().c_str())==-1)
		throw SystemError("unlink failed", errno);
}

list<string> list_files(const Path &path)
{
	list<string> result;
	DIR *dir=opendir(path.str().c_str());
	if(dir)
	{
		while(dirent *de=readdir(dir))
		{
			const char *fn=de->d_name;
			if(fn[0]=='.' && (fn[1]==0 || (fn[1]=='.' && fn[2]==0))) 
				continue;
			result.push_back(fn);
		}
		closedir(dir);
	}

	return result;
}

bool exists(const Path &path)
{
	return access(path.str().c_str(), F_OK)==0;
}

Filename splitext(const string &fn)
{
	Filename result;
	unsigned dot=fn.rfind('.');
	result.base=fn.substr(0, dot);
	if(dot!=string::npos)
		result.ext=fn.substr(dot);
	return result;
}

int fnmatch(const string &pat, const Path &fn)
{
#ifdef WIN32
	return globcasematch(pat, fn.str());
#else
	return ::fnmatch(pat.c_str(), fn.str().c_str(), FNM_PATHNAME);
#endif
}

Path relative(const Path &path, const Path &base)
{
	Path::Iterator i=path.begin();
	Path::Iterator j=base.begin();
	for(; (i!=path.end() && j!=base.end() && *i==*j); ++i,++j);

	Path result;
	for(; j!=base.end(); ++j)
		result/="..";
	for(; i!=path.end(); ++i)
		result/=*i;

	return result;
}

/**
Extracts the basename from the given path.  Same thing as Path::Path(p)[-1],
but faster.
*/
string basename(const std::string &p)
{
	unsigned slash=p.rfind(DIRCHAR);
	if(slash==string::npos)
		return p;
	else
		return p.substr(slash+1);
}

int stat(const Path &fn, struct stat &st)
{
	return ::stat(fn.str().c_str(), &st);
}

struct stat stat(const Path &fn)
{
	struct stat st;
	if(stat(fn, st)==-1)
		throw SystemError("stat failed", errno);
	return st;
}

Path getcwd()
{
	char buf[1024];
	return ::getcwd(buf, sizeof(buf));
}

} // namespace Msp
